@@ -170,9 +170,10 @@ The record is, however, still there:
 
 """
 
-from . import template
+from . import template, types
 from .template import IpfixEncodeError, IpfixDecodeError
 
+import io
 import operator
 import functools
 import struct
@@ -225,8 +226,10 @@ class MessageBuffer:
         
         self.mtu = 65535
         
-        self.add_template_hook = None
-        
+        self.template_record_hook = None
+        self.unknown_data_set_hook = None
+        self.ignored_data_set_hook = None
+
     def __repr__(self):
         if self.cursetid:
             addinf = " (writing set "+str(self.cursetid)+")"
@@ -251,7 +254,7 @@ class MessageBuffer:
         :returns: export time of the last message read/written.
         
         """
-        return datetime.utcfromtimestamp(self.export_epoch)
+        return types._decode_sec(self.export_epoch)
 
     def set_export_time(self, dt=None):
         """
@@ -264,7 +267,7 @@ class MessageBuffer:
         """
         if not dt:
             dt = datetime.utcnow()
-        self.export_epoch = int(dt.timestamp())        
+        self.export_epoch = types._encode_sec(dt)
         self.auto_export_time = False
                 
     def _increment_sequence(self):
@@ -284,7 +287,7 @@ class MessageBuffer:
         while (offset < self.length):
             (setid, setlen) = _sethdr_st.unpack_from(self.mbuf, offset)
             if offset + setlen > self.length:
-                raise IPFIXDecodeError("Set too long for message")
+                raise IpfixDecodeError("Set too long for message")
             self.setlist.append((offset, setid, setlen))
             offset += setlen
     
@@ -331,9 +334,10 @@ class MessageBuffer:
         
         # populate setlist
         self._scan_setlist()
-            
+
     def from_bytes(self, bytes):
-        """Read an IPFIX message from a byte array.
+        """
+        Read an IPFIX message from a byte array.
         
         This populates message header fields and the internal setlist.
         Call for each new message before iterating over records when reading
@@ -343,26 +347,29 @@ class MessageBuffer:
         :raises: IpfixDecodeError
         
         """
-        # make a copy of the byte array
-        self.mbuf[0:len(bytes)] = bytes
+        return self.read_message(io.BytesIO(bytes))
 
-        # parse message header 
-        if (len(bytes) < _msghdr_st.size):
-            raise IpfixDecodeError("Message too short ("+str(len(msghdr)) +")")
+    # def old_from_bytes(self, bytes):
+    #     # make a copy of the byte array
+    #     self.mbuf[0:len(bytes)] = bytes
 
-        (version, self.length, self.sequence, self.export_epoch, self.odid) = \
-                _msghdr_st.unpack_from(self.mbuf, 0)
+    #     # parse message header 
+    #     if (len(bytes) < _msghdr_st.size):
+    #         raise IpfixDecodeError("Message too short ("+str(len(msghdr)) +")")
+
+    #     (version, self.length, self.sequence, self.export_epoch, self.odid) = \
+    #             _msghdr_st.unpack_from(self.mbuf, 0)
         
-        # verify version and length
-        if version != 10:
-            raise IpfixDecodeError("Illegal or unsupported version " + 
-                                   str(version))
+    #     # verify version and length
+    #     if version != 10:
+    #         raise IpfixDecodeError("Illegal or unsupported version " + 
+    #                                str(version))
         
-        if self.length < 20:
-            raise IpfixDecodeError("Illegal message length" + str(self.length))
+    #     if self.length < 20:
+    #         raise IpfixDecodeError("Illegal message length" + str(self.length))
         
-        # populate setlist
-        self._scan_setlist()
+    #     # populate setlist
+    #     self._scan_setlist()
             
     def record_iterator(self, 
                         decode_fn=template.Template.decode_namedict_from, 
@@ -403,25 +410,30 @@ class MessageBuffer:
                         self.accepted_tids.add((self.odid, tmpl.tid))
                     else:
                         self.accepted_tids.discard((self.odid, tmpl.tid))
-                    if self.add_template_hook:
-                        self.add_template_hook(self, tmpl)
+
+                    if self.template_record_hook:
+                        self.template_record_hook(self, tmpl)
                     
             elif setid < 256:
                 warn("skipping illegal set id "+str(setid))
-            elif (self.odid, setid) in self.accepted_tids:
+            else:
                 try:
                     tmpl = self.templates[(self.odid, setid)]
-                    while offset + tmpl.minlength <= setend:
-                        (rec, offset) = decode_fn(tmpl, self.mbuf, offset, 
-                                                  recinf = recinf)
-                        yield rec
-                        self._increment_sequence()
-                except KeyError:
-                    #FIXME provide set buffer for sets without templates
-                    pass
-            else:
-                #FIXME disable sequence checking on skipped sets
-                pass
+                    if (self.odid, setid) in self.accepted_tids:
+                        while offset + tmpl.minlength <= setend:
+                            (rec, offset) = decode_fn(tmpl, self.mbuf, offset, 
+                                                      recinf = recinf)
+                            yield rec
+                            self._increment_sequence()
+                    elif self.ignored_data_set_hook:
+                        # not in accepted tids - ignored data set
+                        self.ignored_data_set_hook(self, tmpl, 
+                                     self.mbuf[offset-_sethdr_st.size:setend])
+                except KeyError as e:
+                    if self.unknown_data_set_hook:                
+                        # KeyError on template lookup - unknown data set
+                        self.unknown_data_set_hook(self, 
+                                     self.mbuf[offset-_sethdr_st.size:setend])
 
     def namedict_iterator(self):
         """
@@ -513,8 +525,8 @@ class MessageBuffer:
         if export:
             self.export_template(tmpl.tid)
         
-        if self.add_template_hook:
-            self.add_template_hook(self, tmpl)
+        if self.template_record_hook:
+            self.template_record_hook(self, tmpl)
     
     def delete_template(self, tid, export=True):
         """
@@ -744,4 +756,4 @@ class MessageBuffer:
         :raises: EndOfMessage
         
         """       
-        self.export_record(rec, template.Template.encode_tuple_to)
\ No newline at end of file
+        self.export_record(rec, template.Template.encode_tuple_to)
